#!/usr/bin/env python3
"""
Return F0 from the Thuillier 2003 lookup table.

Based on the function get_f0.pro

[structure of f0 and lambda] = get_f0(avg_wl=[array of wl])
and avg_wl=[array of wl] calculated 10 nm average for each wl
"""

import numpy as np
import argparse
import sys
import os
from netCDF4 import Dataset
import datetime


def get_f0(wvls=None, obstime=None):
    """
    Load Thuillier2003 netcdf and return f0.

    Defaults to returning the full table. Input obstime to correct for the
    Earth-Sun distance.
    """
    f0file = os.path.join(
        os.path.dirname(__file__), 'data', 'thuillier2003_f0.nc')
    ncfile = Dataset(f0file, 'r')

    wl = ncfile['wavelength'][:]
    f0 = ncfile['irradiance'][:]

    wl_save = wl
    f0_save = f0

    if wvls:
        avg_f0 = np.empty(len(wvls))
        avg_f0[:] = np.nan

        for i in range(len(wvls)):
            idx = np.where((wl >= wvls[i]-5.) & (wl <= wvls[i]+5.))
            if idx:
                avg_f0[i] = np.mean(f0[idx])

        f0_save = avg_f0
        wl_save = wvls

    if obstime:
        f0_save *= correct_esdist(obstime)

    result = (f0_save, wl_save)

    return result


def correct_esdist(obstime):
    """Calculate the Earth-Sun distance correction factor to apply to F0."""
    pi2 = 2 * np.pi
    rjd = obstime.timetuple().tm_yday + obstime.timetuple().tm_hour/24.0
    return np.square(1.0 / (1.00014 - 0.01671*np.cos(1.0*pi2*(0.9856002831*rjd-3.4532868)/360.0)
                            - 0.00014*np.cos(2.0*pi2*(0.9856002831*rjd-3.4532868)/360.0)))


def main():
    """Primary driver for stand-alone version."""
    __version__ = '1.0.0'

    parser = argparse.ArgumentParser(description=(
        'Generates hyperspectral (1nm) or 10nm bandpass averaged '
        'extraterrestrial irradiance spectrum using Thuillier (2002): '))

    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('--output_file', type=str,
                        help='output filename; default is STDOUT')
    parser.add_argument('--wvl', type=float, nargs='+',
                        default=None, help='wavelength(s) to generate')
    parser.add_argument('--datetime', type=str, default=None,
                        help=('ISO8601 date used to correct for '
                              'Earth-Sun distance'))
    parser.add_argument('--verbose', '-v', action='store_true')

    args = parser.parse_args()

    obstime = None
    esdistcorr = 1.0
    if args.datetime:
        obstime = datetime.datetime.fromisoformat(args.datetime)
        esdistcorr = correct_esdist(obstime)

    verbose = args.verbose

    if verbose and args.output_file:
        print("Generating extraterrestrial irradiance spectrum ...")

    (spectrum, wvl) = get_f0(wvls=args.wvl, obstime=obstime)
    outarray = np.column_stack((np.asarray(wvl), np.asarray(spectrum)))

    if args.output_file:
        ofile = open(args.output_file, 'w')
        ofile.write("/begin_header\n")
        ofile.write("! Extraterrestrial irradiance spectrum from Thuillier\n")
        ofile.write("!     Thuillier, G., Hersé, M., Labs, D. et al.\n")
        ofile.write("!     The Solar Spectral Irradiance from 200 to 2400 nm "
                    "as Measured by the SOLSPEC Spectrometer\n")
        ofile.write("!     from the Atlas and Eureca Missions. Solar Physics "
                    "214, 1–22 (2003).\n")
        ofile.write("!     https://doi.org/10.1023/A:1024048429145\n\n")
        if obstime:
            ofile.write("! Observation time: {}\n".format(args.datetime))
            ofile.write("! Earth-Sun distance correction factor: "
                        "{:.7f}\n".format(esdistcorr))
        else:
            ofile.write("! Mean Earth-Sun distance\n")

        if args.wvl:
            ofile.write("! 11nm width bands\n")
        # if args.sensor:
        #     ofile.write("! convolved with the spectral response function "
        #                 "for %s\n" % sensors[args.sensor]['name'])
        ofile.write("/missing=-999\n")
        ofile.write("/delimiter=space\n")
        ofile.write("/fields=wavelength,F0\n")
        ofile.write("/units=nm,mW cm^-2 um^-1\n")
        ofile.write('\n'.join([' '.join(['{:.5f}'.format(value)
                                         for value in row])
                               for row in outarray]))
        ofile.close()
    else:
        print('\n'.join([' '.join(['{:.5f}'.format(value)
              for value in row]) for row in outarray]))



# The following allows the file to be imported without immediately executing.
if __name__ == '__main__':
    sys.exit(main())
