#!/usr/bin/env python3
"""
Ocean Reflectance Model

Generates hyperspectral (1nm) or sensor-specific Rrs (or Lwn) spectra using the
ocean reflectance model described in:
     P. Jeremy Werdell, Sean W. Bailey, Bryan A. Franz, André Morel, and
     Charles R. McClain, "On-orbit vicarious calibration of ocean color sensors
     using an ocean surface reflectance model" Appl. Opt. 46, 5649-5666 (2007),
     https://doi.org/10.1364/AO.46.005649
"""


import math
from netCDF4 import Dataset
import numpy as np
from scipy.interpolate import CubicSpline as spline
from scipy.interpolate import interp1d
import scipy.ndimage.filters as filters

from orm.get_brdf import get_fq
import orm.morel_read as morel_read
from orm.utils import get_data_path, rrs_to_lwn, bandpass_avg_rrs, get_wavelength_array

__version__ = '1.1.0'
verbose = False

# Global variables
wl = get_wavelength_array()
interp_func_q = None
bricdat = None

verbose = False

def rrs_below_to_above(rrs_s, q=3.141529, gordon=False):
    """
    Compute above-water Rrs from sub-surface rrs.

    Uses Lee et al 2002 or Gordon 1988.

    Inputs:
        subsurface remote-sensing reflectance
        Q parameter (for Gordon - defaults to PI)
    Output:
        above-water remote-sensing reflectance
    """
    if gordon:
        rho = 0.021
        rho_bar = 0.043
        r_eu = 0.48
        m = 1.34

        t = ((1. - rho) * (1. - rho_bar)) / (np.power(m,2) * (1. - r_eu * rrs_s))
        rrs_a = t * rrs_s / q

    else:
        rrs_a =  (0.52 * rrs_s) / (1.0 - 1.7 * rrs_s)

    return rrs_a


def get_mm01_bb(chl):
    """
    Retrieve backscattering coefficient for a given chlorophyll concentration.

    Based on Loisel and Morel 1998, Ciotti et al 1999, and
    Morel and Maritorena 2001.
    """
    global wl

    # Backscattering coefficient for seawater
    bbw = 0.0038 * np.power(400./wl, 4.32)

    # Loisel and Morel 1998 bbp function
    bp550 = 0.416 * np.power(chl, 0.766)

    # Morel and Maritorena 2001 bbp with v from Ciotti et al. 1999
    v = 0.768 * np.log10(chl) - 1.

    bbp = (0.002 + 0.01 * (0.5 - 0.25 * np.log10(chl))) * np.power(wl[:,np.newaxis] / 550.,v) * bp550
    bb = bbp + bbw[:, np.newaxis]

    return np.squeeze(bb)


def get_mm01_kd(chl):
    """
    Compute diffuse attenuation coefficient from chlorophyll concentration.

    Based on Morel and Maritorena 2001
    Kd = Kw + X chl^e
    """
    global wl
    kbiofile = get_data_path() / 'mm01_kbio.nc'
    data = Dataset(kbiofile, 'r')
    X = data['X'][:]
    e = data['e'][:]
    kw = data['kw'][:]
    wavelength = data['wavelength'][:]

    k0 = X[:, np.newaxis] * np.power(chl, e[:, np.newaxis]) + kw[:, np.newaxis]

    cs = spline(wavelength, k0)
    kd = cs(wl)

    return np.squeeze(kd)


def get_morel_mud(chl, solz):
    """
    Compute mean cosine for downward flux.

    A. Morel, personal communication, Summer 2006
    """
    mud = morel_read.morel_read(chl, solz, mud=True)

    return mud


def get_morel_fq(chl, solz):
    """
    Retrieve f and Q given chlorophyll and solar zenith angle.

    Morel et al. 2002, Appendix B for use with in-water data
    (senz = 0 and relaz = 0)
    """
    global wl
    global interp_func_q

    if interp_func_q is None:
        print("Initializing q table interpolation")
        interp_func_q = morel_read.morel_q_interp()

    interp_arr = np.meshgrid(solz, chl, wl)
    flat = np.array([m.flatten() for m in interp_arr])
    out_array = interp_func_q(flat.T)
    q = np.moveaxis(out_array.reshape(*interp_arr[0].shape), -1, 0)
    if solz.size > 1:
        q = np.diagonal(q, axis1=1, axis2=2)

    f = morel_read.morel_read(chl, solz, fp=True)

    f = np.squeeze(f)
    q = np.squeeze(q)
    return (f, q)


def get_aw():
    """Get pure water absorption and interpolate to sensor wavelength."""
    specFile = get_data_path() / "water_spectra.nc"
    spec = Dataset(specFile, 'r')
    wvl = spec['wavelength'][:]
    idx = np.where(np.logical_and(wvl >= 300, wvl <= 900))
    aw = spec['aw'][:]
    bw = spec['bw'][:]
    spec.close()

    return aw[idx], bw[idx]/2.


def get_bricaud_aph(chl, normalize=False):
    """
    Get the bricaud phytoplankton absorption shape.

    Bricaud et al., JGR 103, 31033-31044, 1998
    Inputs:
        chl: chlorophyll-a pigment concentration
        norm: logic flag to normalise aph* to 0.055 as per GIOP method

    Ouputs:
        ap: particulate absorption coefficient (Ap Chl^Ep)
        aps: normalised particle absorption coefficient (Ap Chl^(Ep-1))
        aph: phytoplankton absorption coefficient (Aph Chl^Eph)
        aphs: normalised phytoplankton absorption coefficient (Aph Chl^(Eph-1))
    """
    global wl
    global bricdat

    # Load files once
    if bricdat is None:
        # bricaud_1998_aph.txt order of columns: wl,Ap,Ep,Aph,Eph
        bricFile = get_data_path() / "aph_bricaud_1998.txt"
        print("Loading {}".format(str(bricFile)))
        bricdat = np.genfromtxt(bricFile, delimiter=',', skip_header=12)
    chl = np.atleast_1d(chl)
    ap0 = bricdat[:, 1] * np.power(chl[:, np.newaxis], bricdat[:, 2])
    aps0 = bricdat[:, 1] * np.power(chl[:, np.newaxis], (bricdat[:, 2] - 1))
    aph0 = bricdat[:, 3] * np.power(chl[:, np.newaxis], bricdat[:, 4])
    aphs0 = bricdat[:, 3] * np.power(chl[:, np.newaxis], (bricdat[:, 4] - 1))

    # Normalize aph
    if normalize:
        idx442 = np.where(bricdat[:, 0] == 442)[0][0]
        aphs442 = aphs0[:, idx442]
        aphs0 = aphs0 * 0.055 / aphs442[:, np.newaxis]

    # Interpolate to sensor wavelength
    ap = interp1d(bricdat[:, 0], ap0, bounds_error=False,
                  kind='cubic', fill_value=0.0)
    aps = interp1d(bricdat[:, 0], aps0, bounds_error=False,
                   kind='cubic', fill_value=0.0)
    aph = interp1d(bricdat[:, 0], aph0, bounds_error=False,
                   kind='cubic', fill_value=0.0)
    aphs = interp1d(bricdat[:, 0], aphs0,
                    bounds_error=False, kind='cubic', fill_value=0.0)

    return ap(wl), aps(wl), aph(wl), aphs(wl)


def extrap_aph(chl, normalize=False):
    """
    Extrapolate aphStar into the UV and NIR.

    Inputs:
       chl: chlorophyll-a pigment concentration (float)

    Ouputs:
       aphStar:  extrapolated normalised phytoplankton absorption shape
    """
    global wl
    ap, aps, aph, aphs = get_bricaud_aph(chl, normalize=normalize)

    # very, very ad-hoc extrapolation from 400 down to 350 nm, then spectrally
    # flat from from 350 nm to shorter wavelengths
    i350 = np.where(wl == 350)[0].item()
    i400 = np.where(wl == 400)[0].item()
    i405 = np.where(wl == 405)[0].item()
    i410 = np.where(wl == 410)[0].item()
    i443 = np.where(wl == 443)[0].item()
    i700 = np.where(wl == 700)[0].item()

    # The 0.5 factor here is arbitrary to stop extrapolations going negative
    aph_slope = 0.5*(aphs[:, i443] - aphs[:, i410]) / (443 - 410)
    aph_int = aphs[:, i443] - aph_slope*443
    lamUV = wl[i350:i405]
    aphUV = aph_slope*lamUV[:, np.newaxis] + aph_int
    delta_aphs = aphUV[-1] - aphs[:, i405]
    aphUV = aphUV - delta_aphs

    aphs[:, i350:i405] = np.transpose(aphUV)
    aphs[:, 0:i350] = (aphs[:, i350])[:, np.newaxis]

    # Wavelengths longer than 700 nm extrapolated using a smooth exponential
    # decay curve. Curve shape derived from selection of SeaBASS measurements
    p = np.array((-0.0789811, 55.12363434))
    aph700Nir = np.exp(p[1])*np.exp(wl[i700:]*p[0])

    aphs[:, i700:] = aph700Nir*(aphs[:, i700])[:, np.newaxis]

    aphUV = np.array([0.8168912, 0.81677942, 0.81650336, 0.81598752,
                      0.81518127, 0.81410382, 0.81286164, 0.81162225,
                      0.81055652, 0.80978012, 0.80932177, 0.80912699,
                      0.80909137, 0.80910478, 0.8091019, 0.80910205,
                      0.80921815, 0.80962099, 0.8104629, 0.81180237,
                      0.81357246, 0.815613, 0.81774113, 0.81981493,
                      0.82175759, 0.82354585, 0.8251863, 0.82670197,
                      0.82812943, 0.82951537, 0.83089821, 0.8322774,
                      0.8335878, 0.8346934, 0.83540505, 0.83551171,
                      0.83480672, 0.83310557, 0.83026425, 0.82620653,
                      0.82095056, 0.81461136, 0.80736381, 0.79937762,
                      0.79076349, 0.78156979, 0.77183674, 0.76168252,
                      0.75136854, 0.74130173, 0.73196185, 0.72378628,
                      0.71706382, 0.711885, 0.70816165, 0.70569461,
                      0.70425134, 0.70362551, 0.70367264, 0.70432556,
                      0.70559371, 0.70753984, 0.71023409, 0.71370017,
                      0.71788149, 0.72265037, 0.72785336, 0.7333654,
                      0.73912123, 0.74511257, 0.75136361, 0.75790804,
                      0.76477631, 0.7719909, 0.77955978, 0.78746745,
                      0.7956687, 0.80409496, 0.81267063, 0.82132997,
                      0.83002155, 0.83869668, 0.84729066, 0.85571288,
                      0.86385694, 0.87162939, 0.87898592, 0.88595694,
                      0.89265189, 0.8992391, 0.90590968, 0.91283903,
                      0.9201604, 0.92795851, 0.93628156, 0.9451633,
                      0.9546435, 0.96478, 0.9756518, 0.98735552, 1.])

    lamAphUV = np.linspace(300, 400., 101)

    aphUVI = interp1d(lamAphUV, aphUV, bounds_error=False,
                      kind='cubic', fill_value=0.0)
    aphUV_temp = aphUVI(wl[0:i400+1])

    aphUV = aphUV_temp*(aphs[:, i400])[:, np.newaxis]

    aphs[:, 0:i400+1] = aphUV

    # smooth a little using a gaussain convolution filter
    # For this excercise, we are simulating Rrs for testing processing
    # capabilities and not algorithm skill
    aphs = filters.gaussian_filter1d(aphs, sigma=2)

    # normalise to 1 at 443 nm
    aphStar = aphs/(aphs[:, i443])[:, np.newaxis]

    return aphStar.T


def morel_rrs(chl, solz=None, wvl=None, bandpass=11,
              lwn=False, mm01iter=True, use_fq=False):
    """
    Calculate Rrs given chlorophyll and solar zenith angle.

    Based mostly on Morel and Maritorena 2001
    Rrs = trans * (f / Q) * {bb / (a + bb)}, where (a + bb) ~ mud * Kd
    """
    global wl
    global verbose

    # Default solar zenith angle is 30-degrees
    if solz is None:
        solz = np.atleast_1d(30.)
    else:
        solz = np.atleast_1d(solz)

    chl = np.atleast_1d(chl)

    # Backscattering coefficient
    bb = get_mm01_bb(chl)

    # Diffuse attenuation coefficient
    kd = get_mm01_kd(chl)

    # Mean cosine for downward flux
    mud = get_morel_mud(chl, solz)

    # f and Q
    (f, q) = get_morel_fq(chl, solz)
    if use_fq:
        fq = get_fq(solz, 0., 0., chl, wvl=wvl)
        f = fq * q

        # Morel and Maritorena 2001, Morel et al. 2007 iteration
    u2 = mud
    r0 = f * bb / (u2 * kd + bb)

    if mm01iter:
        # Iterate 3 times
        for i in range(3):
            u2 = mud * (1. - r0) / (1. + mud * r0 / 0.4)
            atot = 0.962 * kd * mud * (1. - r0 / f)
            r0 = f * bb / (atot + bb)
    else:
        # Iterate until convergence
        last_r0 = r0 + 2.
        iterations = 0
        close_enough = False
        while close_enough is False:
            iterations += 1
            u2 = mud * (1. - r0) / (1. + mud * r0 / 0.4)
            atot = 0.962 * kd * mud * (1. - r0 / f)
            r0 = f * bb / (atot + bb)
            close_enough = True
            for i in range(len(r0)):
                if not math.isclose(r0[i], last_r0[i], rel_tol=1e-07):
                    last_r0[i] = r0[i]
                    close_enough = False
                    continue
            if close_enough:
                break

        if verbose:
            print("MM01 iterations: %d" % iterations)

    # Transmission across the air-sea interface
    rrs = rrs_below_to_above(r0, q, gordon=True)

    # lwn instead of rrs
    if lwn:
        rrs = rrs_to_lwn(rrs)

        # Bandpass averages centered on discrete lambda
    if wvl is not None:
        rrs = bandpass_avg_rrs(rrs, wvl, bandpass=bandpass)

    return rrs


def gordon_rrs(adg, aph, bbp, bbp_s, chl, lwn=False, wvl=None, bandpass=11):
    """
    Calculate Rrs given chlorophyll and IOPs.

    Uses the Gordon quadratic formulation.
    """
    global wl
    aw, bbw = get_aw()
    g0 = 0.0949
    g1 = 0.0794

    # Model adgStar spectra
    madgStar = np.exp(-0.018 * (wl[:, np.newaxis]-443.))

    # Get extrapolated aphStar spectra
    maphStar = extrap_aph(chl, normalize=False)

    # Get bbpStar spectra
    mbbpStar = np.power((443./wl[:, np.newaxis]), bbp_s)

    # Compute total backscattering
    bbtot = np.squeeze(bbp * mbbpStar + bbw[:, np.newaxis])
    # bbtot = get_mm01_bb(chl)

    # Compute total absorption
    atot = np.squeeze(aw[:, np.newaxis] + adg * madgStar + aph * maphStar)

    kappa = atot + bbtot
    u = bbtot / kappa
    rrs = np.squeeze((g0 + g1 * u) * u)

    # Transmission across the air-sea interface
    Rrs = rrs_below_to_above(rrs)

    # lwn instead of rrs
    if lwn:
        Rrs = rrs_to_lwn(Rrs)

    # Bandpass averages centered on discrete lambda
    if wvl is not None:
        Rrs = bandpass_avg_rrs(Rrs, wvl, bandpass=bandpass)

    return Rrs