"""Modules that read in Morel Tables."""
import argparse
import sys
import numpy as np
from netCDF4 import Dataset
from scipy.interpolate import RegularGridInterpolator as rgi
import os

wl = np.linspace(300, 900., 601)
interp_func_mud = None
interp_func_fn = None
interp_func_fp = None

def get_script_path():
    """Return path to orm_morel table."""
    return os.path.join(os.path.dirname(__file__),'data')

def morel_table_interp(mud=False, fn=False, fp=False):
    """Read Morel f, fp, and mud tables and interpolate."""
    mudfile = os.path.join(get_script_path(), 'morel_mud_f_fp.nc')
    ncfile = Dataset(mudfile, 'r')
    data = None
    if mud:
        data = ncfile['mud'][:, :, :]
    elif fn:
        data = ncfile['f'][:, :, :]
    elif fp:
        data = ncfile['fp'][:, :, :]

    if data is None:
        print, ''
        print, 'Error: please indicate mud, f, or fp'
        print, ''
        sys.exit(-1)

    grdwvl = ncfile['wavelength'][:]
    grdchl = ncfile['chlorophyll'][:]
    grdsolz = ncfile['solz'][:]

    interp_func = rgi((grdsolz, grdwvl, grdchl), data,
                      fill_value=None, bounds_error=False)

    return interp_func

def morel_q_interp():
    """Read Morel f/Q table and interpolate."""
    fqfile = os.path.join(get_script_path(), 'morel_fq_appb.nc')

    morel_fq_appb = Dataset(fqfile, 'r')

    grdwvl = morel_fq_appb['wavelength'][:]
    grdchl = morel_fq_appb['chlorophyll'][:]
    q0 = morel_fq_appb['q0'][:]
    sq = morel_fq_appb['sq'][:]
    grdsolz = np.linspace(0., 60., num=30)

    z = 1. - np.cos(np.deg2rad(grdsolz))
    q1 = np.empty([grdsolz.size, grdchl.size, grdwvl.size])
    for i, solz in enumerate(grdsolz):
        q1[i, :, :] = q0 + sq * z[i]

    return rgi((grdsolz, grdchl, grdwvl), q1,
               fill_value=None, bounds_error=False)


def morel_read(chl, solz, mud=False, fn=False, fp=False):
    """Standalone version, reads Morel tables and interpolates."""
    global interp_func_mud
    global interp_func_fn
    global interp_func_fp
    global wl
    solz = np.atleast_1d(solz)
    chl = np.atleast_1d(chl)

    if mud:
        if interp_func_mud is None:
            interp_func_mud = morel_table_interp(mud=True)
            print("Initializing mean cosine of downwelling irradiance "
                  "interpolation function")
        interp_func = interp_func_mud
    elif fn:
        if interp_func_fn is None:
            interp_func_fn = morel_table_interp(fn=True)
            print("Initializing f interpolation function")
        interp_func = interp_func_fn
    elif fp:
        if interp_func_fp is None:
            interp_func_fp = morel_table_interp(fp=True)
            print("Initializing f-prime interpolation function")
        interp_func = interp_func_fp
    else:
        print, ''
        print, 'Error: please indicate mud, f, or fp'
        print, ''
        sys.exit(-1)

    interp_arr = np.meshgrid(solz, wl, chl)
    flat = np.array([m.flatten() for m in interp_arr])
    out_array = interp_func(flat.T)
    result = np.squeeze(out_array.reshape(*interp_arr[0].shape))

    if solz.size > 1:
        result = np.diagonal(result, axis1=1, axis2=2)

    return result
    
def main():
    """Primary driver for stand-alone version."""
    __version__ = '1.0.0'

    parser = argparse.ArgumentParser(description='Read Morel tables')

    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('chl', type=float, default=None,
                        help='input chlorophyll')
    parser.add_argument('solz', type=float, default=None,
                        help='input solar zenith angle')
    parser.add_argument('--table','-t', type=str,
                        choices=['mud',
                                 'fn',
                                 'fp'],
                        default=None, help='Morel table to interpolate')
    args = parser.parse_args()

    global wl

    tables = {'mud':False,
              'fp':False,
              'fn':False}
    for table in tables.keys():
        if table == args.table:
            tables[table]=True

    spectrum = morel_read(args.chl, args.solz, mud=tables['mud'], fp=tables['fp'], fn=tables['fn'])
    outarray = np.column_stack((np.asarray(wl), np.asarray(spectrum)))
    print('\n'.join([' '.join(['{:.6f}'.format(value)
                        for value in row])
                            for row in outarray]))

# The following allows the file to be imported without immediately executing.
if __name__ == '__main__':
    sys.exit(main())