#!/usr/bin/env python3
"""
Ocean Reflectance Model Script.

Generates hyperspectral (1nm) or sensor-specific Rrs (or Lwn) spectra using the
ocean reflectance model described in:
     P. Jeremy Werdell, Sean W. Bailey, Bryan A. Franz, André Morel, and
     Charles R. McClain, "On-orbit vicarious calibration of ocean color sensors
     using an ocean surface reflectance model" Appl. Opt. 46, 5649-5666 (2007),
     https://doi.org/10.1364/AO.46.005649
"""

import argparse
import numpy as np
import sys

from orm.model import morel_rrs, gordon_rrs
from orm.utils import convolve_rsr

__version__ = '1.1.0'
verbose = False

# Global variables
wl = np.linspace(300, 900., 601)
interp_func_q = None
bricdat = None

rsr_wvl = None
rsr = None
rsr_bands = None
rsr_sensor = None

verbose = False


def main():
    """Primary driver for stand-alone version."""
    global verbose
    global wl
    global sensors

    np.set_printoptions(precision=6)

    parser = argparse.ArgumentParser(description=(
        'Generates hyperspectral (1nm) or sensor-specific Rrs (or Lwn) '
        'spectra using the ocean reflectance model described in: P. Jeremy '
        'Werdell, Sean W. Bailey, Bryan A. Franz, André Morel, and Charles R. '
        'McClain, "On-orbit vicarious calibration of ocean color sensors '
        'using an ocean surface reflectance model," Appl. Opt. 46, 5649-5666 '
        '(2007), https://doi.org/10.1364/AO.46.005649'))

    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + __version__)
    parser.add_argument('--sensor', type=str,
                        choices=['modisa',
                                 'modist',
                                 'viirsn',
                                 'viirsj1',
                                 'seawifs',
                                 'meris',
                                 'olcis3a',
                                 'olcis3b',
                                 'octs',
                                 'czcs'],
                        default=None, help='Sensor to mimic')
    parser.add_argument('--output_file', type=str,
                        help='output netCDF LUT filename; default is STDOUT')
    parser.add_argument('chl', nargs='+', type=float,
                        default=0.05, help='Chlorophyll concentration')
    parser.add_argument('--adg', nargs='?', type=float, default=0.005,
                        help=('Detrital and Gelbstoff absorption coefficient '
                              'at 443nm'))
    parser.add_argument('--aph', nargs='?', type=float, default=0.005,
                        help='Phytoplankton absorption coefficient at 443nm')
    parser.add_argument('--bbp', nargs='?', type=float, default=0.0015,
                        help='Particulate backscattering coefficient at 443nm')
    parser.add_argument('--bbp_s', nargs='?', type=float,
                        default=2.0, help='Particulate backscattering slope')
    parser.add_argument('--solz', nargs='+', type=float, default=30.,
                        help=('Solar zenith angle. If provided, it should '
                              'either be a single value or an array matching '
                              'the size of the input chlorophyll array. The '
                              'default is 30 degrees'))
    parser.add_argument('--wvl', type=float, nargs='+',
                        default=None, help='Wavelength(s) to generate')
    parser.add_argument('--bandpass', type=int, default=11,
                        help='Bandpass width of wavelength(s) to generate')
    parser.add_argument('--lwn', action='store_true',
                        help="output normalized Lw instead of Rrs")
    parser.add_argument('--disable_iter_limit', action='store_false',
                        help="allow MM01 iteration to run to convergence")
    parser.add_argument('--gordon', '-g', action='store_true',
                        help=('Use the Gordon quadratic formulation with '
                              'input IOPs'))
    parser.add_argument('--verbose', '-v', action='store_true')

    args = parser.parse_args()

    verbose = args.verbose
    if args.sensor and args.wvl:
        print("Can't do both sensor and an input wavelength set")
        sys.exit(1)

    wvl = wl
    if args.wvl:
        wvl = args.wvl
        if verbose:
            print("Generating %dnm bandpass spectrum for input wavelengths" %
                  args.bandpass)

    if verbose and args.output_file:
        print("Generating Rrs spectrum %s ..." % args.output_file)

    otype = 'rrs'
    units = '1/sr'
    if args.lwn:
        otype = 'nLw'
        units = 'uW/cm^2/nm'

    solz = np.atleast_1d(args.solz)
    spectrum = None
    if args.gordon:
        spectrum = gordon_rrs(args.adg, args.aph, args.bbp, args.bbp_s,
                              args.chl, bandpass=args.bandpass, lwn=args.lwn)

    else:
        spectrum = morel_rrs(args.chl, solz=solz, wvl=args.wvl,
                             bandpass=args.bandpass, lwn=args.lwn,
                             mm01iter=args.disable_iter_limit)

    # Convolve with sensor RSR, if desired...
    if args.sensor:
        if verbose:
            print(("Generating spectrum based on the spectral response "
                   "functions for %s") % args.sensor)
        (wvl, spectrum) = convolve_rsr(args.sensor, spectrum,verbose=verbose)

    outarray = np.column_stack((np.asarray(wvl), np.asarray(spectrum)))

    if args.output_file:
        if np.asanyarray(args.chl).size > 1:
            chlstr = ['{:f}'.format(x) for x in args.chl]
        else:
            chlstr = '{:f}'.format(args.chl)
        if np.asarray(args.solz).size > 1:
            solzstr = ['{:f}'.format(x) for x in args.solz]
        else:
            solzstr = '{:f}'.format(args.solz)

        ofile = open(args.output_file, 'w')
        ofile.write("/begin_header\n")
        ofile.write("! Output of Ocean Reflectance Model\n")
        ofile.write("! chlorophyll: %s\n" % chlstr)
        ofile.write("! solar zenith angle: %s\n" % solzstr)
        if args.bandpass:
            ofile.write("! band width: %d\n" % args.bandpass)
        if args.sensor:
            ofile.write(("! convolved with the spectral response function "
                         "for %s\n") % sensors[args.sensor]['name'])
        ofile.write("/missing=-999\n")
        ofile.write("/delimiter=space\n")
        ofile.write("/fields=wavelength,%s\n" % otype)
        ofile.write("/units=nm,%s\n" % units)
        ofile.write('\n'.join([' '.join(['{:.6f}'.format(value)
                                         for value in row])
                               for row in outarray]))
        ofile.close()
    else:
        print('\n'.join([' '.join(['{:.6f}'.format(value)
              for value in row]) for row in outarray]))


# The following allows the file to be imported without immediately executing.
# if __name__ == '__main__':
#     sys.exit(main())
