import collections
from netCDF4 import Dataset
import numpy as np
import os
from pathlib import Path
import requests
from scipy.interpolate import interp1d
import sys

import orm.get_f0 as get_f0

rsr_wvl = None
rsr = None
rsr_bands = None
rsr_sensor = None

sensors = collections.defaultdict(dict)

sensors['modisa']['name'] = "MODIS-Aqua"
sensors['modisa']['rsr'] = "aqua_modis_RSR.nc"
sensors['modist']['name'] = "MODIS-Terra"
sensors['modist']['rsr'] = "terra_modis_RSR.nc"
sensors['seawifs']['name'] = "SeaWiFS"
sensors['seawifs']['rsr'] = "orbview-2_seawifs_RSR.nc"
sensors['meris']['name'] = "MERIS"
sensors['meris']['rsr'] = "envisat_meris_RSR.nc"
sensors['olcis3a']['name'] = "OLCI-S3A"
sensors['olcis3a']['rsr'] = "sentinel-3a_olci_RSR.nc"
sensors['olcis3b']['name'] = "OLCI-S3B"
sensors['olcis3b']['rsr'] = "sentinel-3b_olci_RSR.nc"
sensors['octs']['name'] = "OCTS"
sensors['octs']['rsr'] = "adeos_octs_RSR.nc"
sensors['oci']['name'] = "OCI"
sensors['oci']['rsr'] = "PACE_OCI_RSRs.nc"
sensors['czcs']['name'] = "CZCS"
sensors['czcs']['rsr'] = "nimbus-7_czcs_RSR.nc"
sensors['viirsn']['name'] = "VIIRS-SNPP"
sensors['viirsn']['rsr'] = "suomi-npp_viirs_RSR.nc"
sensors['viirsj1']['name'] = "VIIRS-JPSS1"
sensors['viirsj1']['rsr'] = "jpss-1_viirs_RSR.nc"
sensors['viirsj2']['name'] = "VIIRS-JPSS1"
sensors['viirsj2']['rsr'] = "jpss-2_viirs_RSR.nc"
sensors['hawkeye']['name'] = "Hawkeye"
sensors['hawkeye']['rsr'] = "seahawk1_hawkeye_RSR.nc"
sensors['hico']['name'] = "HICO"
sensors['hico']['rsr'] = "iss_hico_RSR.nc"
sensors['glimr']['name'] = "GLIMR"
sensors['glimr']['rsr'] = "glimr_RSR.nc"

# aviris_RSR.nc
# coms_goci_RSR.nc
# gcom_c_sgli_RSR.nc
# irs-p4_ocm_RSR.nc
# kompsat_osmi_RSR.nc
# landsat-8_oli_RSR.nc
# oceansat-2_ocm-2_RSR.nc
# sentinel-2a_msi_RSR.nc
# sentinel-2b_msi_RSR.nc
# terra_misr_RSR.nc

def get_wavelength_array():
    return  np.linspace(300, 900., 601)

def get_data_path(subdir=None,envroot=None):
    """Get path to OCDATAROOT and subdirectories."""
    scriptpath = Path(os.path.dirname(__file__)) / 'data'
    if envroot:
        if os.getenv(envroot):
            scriptpath = Path(os.getenv(envroot))
        else:
            sys.exit("${} needs to be defined".format(envroot))

    if subdir:
        scriptpath = scriptpath / subdir

    if not Path.exists(scriptpath):
        err_msg = "Cannot access %s" % scriptpath
        sys.exit("Cannot access {scriptpath}".format(scriptpath=str(scriptpath)))


    return scriptpath

def download_rsrfile(remotefile,localfile):
    with requests.get(remotefile, stream=True) as r:
        r.raise_for_status()
        with open(localfile, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

def read_sensor_RSR (sensor, verbose=False):
    """
    Read a sensor spectral response (RSR) from the OCW through https.

    Files can be found at oceancolor.gsfc.nasa.gov/resources/docs/rsr_tables/
    Input:
        Sensor name (string)
    Output:
        normalized Rsr (wavelength, sensor band)
        wavelengths of spectral response function
        sensor band labels
    """
    rsrfile = None
    try:
        rsrfile = sensors[sensor]['rsr']
    except:
        print("Unkown sensor")
        sys.exit()
        
    webpath = 'https://oceancolor.gsfc.nasa.gov/images/rsr/' + rsrfile
    datapath = get_data_path() / rsrfile

    if not datapath.exists():
        if verbose:
            print("Local RSR file not found, retrieving: {}".format(sensors[sensor]['rsr']))
        try:
            download_rsrfile(webpath,datapath)
        except:
            print("Unable to retrieve RSR file from:\n\t{}".format(webpath))
            sys.exit()

    rsr_file = Dataset(datapath,'r')
    Rsr = rsr_file['RSR'][:]
    wavelength = rsr_file['wavelength'][:]
    bands = rsr_file['bands'][:]
    rsr_file.close()

    return wavelength, Rsr.T, bands

def convolve_rsr(sensor, spectrum, spectrum_wvl=None, verbose=False):
    """
    Convolve Rrs with a spectral response function.
    input spectrum assumed to be 1nm resoltion from 300 to 900nm unless
    spectrum_wvl is provided 
    """
    global rsr_wvl
    global rsr
    global rsr_bands
    global rsr_sensor
    global sensors
    wl = get_wavelength_array()
    # if provided spectrum is not 601 elements between 300 and 900nm
    if spectrum_wvl is not None:
        wl = spectrum_wvl
        # spline the spectrum to 1nm resolution, if necesssary
        if (np.amax(wl) - np.amin(wl)) / len(wl) != 1:
            spectrum_interp = interp1d(wl, spectrum, kind='cubic', axis=0)
            wl = np.linspace(np.amin(wl), np.amax(wl) , len(wl)+1)
            spectrum = spectrum_interp(wl)

    if rsr is None or sensor != rsr_sensor:
        print("loading spectral response function for {}".format(
            sensors[sensor]['name']))
        rsr_sensor = sensor
        rsr_wvl, rsr, rsr_bands = read_sensor_RSR(rsr_sensor,verbose=verbose)

    # Get bands for the sensor and limit to the range of the global wl set

    bands = rsr_bands
    bidx = np.where((bands >= np.amin(wl)) & (bands <= np.amax(wl)))
    bands = bands[bidx]

    sidx = np.where((wl >= np.amin(rsr_wvl)) & (wl <= np.amax(rsr_wvl)))
    spectrum = spectrum[sidx]

    # Handle cases where input RSR is not at 1nm intervals
    if (np.amax(rsr_wvl) - np.amin(rsr_wvl)) / len(rsr_wvl) != 1:
        rsr_interp = interp1d(rsr_wvl, rsr, kind='cubic', axis=0)
        rsr = rsr_interp(wl[sidx])
        rsr_wvl = wl[sidx]

    rsr_spectrum = np.empty(len(bands))
    idx = np.where((rsr_wvl >= np.amin(wl)) & (rsr_wvl <= np.amax(wl)))
    rsr = rsr[idx]

    if len(spectrum.shape) == 2:
        rsr_spectrum = np.empty([len(bands), spectrum.shape[1]])
        for j in range(spectrum.shape[1]):
            for i, _ in enumerate(bands):
                rsr_spectrum[i][j] = (np.trapz(spectrum[:, j]
                                               * rsr.transpose()[i, :])
                                      / np.trapz(rsr.transpose()[i, :]))
    else:
        for i, _ in enumerate(bands):
            rsr_spectrum[i] = (np.trapz(spectrum*rsr.transpose()[i, :])
                               / np.trapz(rsr.transpose()[i, :]))
    return bands, rsr_spectrum

def rrs_to_lwn(rrs):
    """Convert rrs to Lwn."""
    wl = get_wavelength_array()

    f0, f0wl = get_f0.get_f0()
    widx = np.where((f0wl >= np.amin(wl)) & (f0wl <= np.amax(wl)))
    f0 = f0[widx]
    if len(rrs.shape) == 2:
        f0 = np.expand_dims(f0, axis=-1)

    return rrs * f0

def bandpass_avg_rrs(rrs, wvl, bandpass=11):
    """Calculate bandpass averaged Rrs centered on discrete wavelengths."""
    wl = get_wavelength_array()
    dim = [len(wvl), 1]
    if len(rrs.shape) == 2:
        dim[1] = rrs.shape[1]
    discrete = np.empty(dim)
    discrete[:] = np.nan
    delta = bandpass / 2.
    for i in range(len(wvl)):
        v = np.where((wl >= wvl[i]-delta) & (wl <= wvl[i]+delta))
        if v:
            if len(rrs.shape) == 2:
                discrete[i, :] = np.mean(rrs[v, :], axis=1)
            else:
                discrete[i, :] = np.mean(rrs[v])

    return np.squeeze(discrete)
